// Needs fixed prulib 
// http://e2e.ti.com/support/arm/sitara_arm/f/791/t/239735.aspx
// https://github.com/beagleboard/am335x_pru_package/issues/3

// This module is the routines needed to emulate a MFM disk. The emulated
// disk data is the raw clock and data MFM bit stream for each track.
// The ARM handles reading and writing a cylinder of track data to DDR 
// memory buffer. PRU 1 converts the MFM bits to Pulse Width Modulation (PWM)
// words PRU 0 uses to generate the MFM bitstream. PRU 1 also converts the
// MFM delta transitions times from PRU 0 to the MFM bitstream and stores
// it back in DDR memory. PRU 0 handles outputing the PWM words to the PWM
// generator, reading the delta words from the hardware and handling the
// various drive control lines. If the host seeks PRU 0 interrupts the ARM
// to let it fetch the next cylinder.
//

// Copyright 2014 David Gesswein.
// This file is part of MFM disk utilities.
//
// 09/06/14 DJG Fixed deadlock between shutting down PRU 0 and seeking
//    to next cylinder
//
// MFM disk utilities is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// MFM disk utilities is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with MFM disk utilities.  If not, see <http://www.gnu.org/licenses/>.
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <inttypes.h>
#include <signal.h>
#include <pthread.h>

// MFM program include files
#include <prussdrv.h>
#include <pruss_intc_mapping.h>

#include "msg.h"
#include "pru_setup.h"
#include "emu_tran_file.h"
#include "parse_cmdline.h"

#include "cmd.h"

#define ARRAYSIZE(x) (sizeof(x) / sizeof(x[0]))

// Real disk read/write thread
pthread_t read_thread;

// Find a new home for this
//
// This routine will either set the CPU speed to maximum or restore it
// back to the previous setting. The normal governor didn't boost the
// processor sufficiently to prevent getting behind in processing.
//
// restore: 0 to set speed to maximum, 1 to restore value from previous
//    call
int set_restore_max_cpu_speed(int restore) {
   FILE *file;
   static char governor[100];
   static int freq_changed = 0;
   char maxfreq[100];

   if (!restore) {
      file = fopen("/sys/devices/system/cpu/cpu0/cpufreq/scaling_governor", "r");
      if (file == NULL) {
         return -1;
      }
      if (fscanf(file, "%100s", governor) < 1) {
         return -1;
      }
      file = fopen("/sys/devices/system/cpu/cpu0/cpufreq/scaling_governor", "w");
      if (file == NULL) {
         return -1;
      }
      if (fprintf(file,"userspace") < 1) {
         return -1;
      }
      fclose(file);
      file = fopen("/sys/devices/system/cpu/cpu0/cpufreq/cpuinfo_max_freq", "r");
      if (file == NULL) {
         return -1;
      }
      if (fscanf(file, "%100s", maxfreq) < 1) {
         return -1;
      }
      fclose(file);

      file = fopen("/sys/devices/system/cpu/cpu0/cpufreq/scaling_setspeed", "w");
      if (file == NULL) {
         return -1;
      }
      if (fprintf(file,maxfreq) < 1) {
         return -1;
      }
      fclose(file);
      freq_changed = 1;
   } else {
      if (freq_changed) {
         file = fopen("/sys/devices/system/cpu/cpu0/cpufreq/scaling_governor", "w");
         if (file == NULL) {
            return -1;
         }
         if (fprintf(file,governor) < 1) {
            return -1;
         }
         fclose(file);
      }
      freq_changed = 0;
   }
   return 0;
}


// This routine is for cleaning up when shutting down. Its installed as an
// atexit routine and called by SIGINT handler.
void shutdown(void)
{
   static int called = 0;

   if (called)
      return;
   called = 1;

   set_restore_max_cpu_speed(1);

   pru_write_word(MEM_PRU0_DATA,PRU0_EXIT, 1);

   pthread_join(read_thread, NULL);
   pru_shutdown();
}

// SIGINT/ control-c handler. Call our shutdown and exit
void shutdown_signal(void)
{
   shutdown();
   exit(1);
}

// This thread writes a cylinder of data to the shared DDR memory for the
// PRU's. It waits for an interrupt, writes the buffer back to disk if dirty,
// then reads the next cylinder of data
// TODO: Buffer write data to prevent hanging if write blocks. Speed up
// memory copy.
static void *emu_proc(void *arg)
{
   DRIVE_PARAMS *drive_params = arg;
   int cyl[MAX_DRIVES];
   int new_cyl[MAX_DRIVES];
   uint32_t *data[MAX_DRIVES];
   uint32_t b;
   double seek_time, max_seek_time = 0;
   int dirty;
   EMU_FILE_INFO emu_file_info[MAX_DRIVES];
   int cyl_size[MAX_DRIVES];
   int i;
   int done = 0;

   // Heads must be zero for unused drives
   for (i = 0; i < MAX_DRIVES; i++) {
      pru_write_word(MEM_PRU0_DATA,PRU0_DRIVE0_NUM_HEAD + i*PRU_WORD_SIZE_BYTES,
            0);
   }

   // Send parameters for drives emulating to PRU
   for (i = 0; i < drive_params->num_drives; i++) {
      cyl[i] = 0;
      new_cyl[i] = 0;
      drive_params->fd[i] = emu_file_read_header(drive_params->filename[i],
            &emu_file_info[i], 1);

      cyl_size[i] = (emu_file_info[i].track_data_size_bytes +
             emu_file_info[i].track_header_size_bytes) * emu_file_info[i].num_head;
      data[i] = malloc(cyl_size[i]);
      if (data[i] == NULL) {
         msg(MSG_FATAL, "Cylinder buffer malloc failed, size %d\n", cyl_size[i]);
         exit(1);
      }
      msg(MSG_INFO,"Drive %d num cyl %d num head %d track len %d\n", i,
        emu_file_info[i].num_cyl, emu_file_info[i].num_head,
        emu_file_info[i].track_data_size_bytes);

      pru_write_word(MEM_PRU0_DATA,PRU0_DRIVE0_NUM_HEAD + i*PRU_WORD_SIZE_BYTES,
            emu_file_info[i].num_head);
      pru_write_word(MEM_PRU0_DATA,PRU0_DRIVE0_NUM_CYL + i*PRU_WORD_SIZE_BYTES,
            emu_file_info[i].num_cyl);
      pru_write_word(MEM_PRU1_DATA,PRU1_DRIVE0_TRACK_HEADER_BYTES +
            i*PRU_WORD_SIZE_BYTES, emu_file_info[i].track_header_size_bytes);
      pru_write_word(MEM_PRU1_DATA,PRU1_DRIVE0_TRACK_DATA_BYTES +
            i*PRU_WORD_SIZE_BYTES, emu_file_info[i].track_data_size_bytes);

      // Transfer cylinder 0 data
      emu_file_read_cyl(drive_params->fd[i], 0, data[i],  cyl_size[i]);
      pru_write_mem(MEM_DDR, data[i], cyl_size[i], i*DDR_DRIVE_BUFFER_MAX_SIZE);
   }


   while (!done) {
      for (i = 0; i < drive_params->num_drives; i++) {
         if (new_cyl[i] != cyl[i]) {
            cyl[i] = new_cyl[i];
            emu_file_read_cyl(drive_params->fd[i], cyl[i], data[i],  cyl_size[i]);
            pru_write_mem(MEM_DDR, data[i], cyl_size[i], i*DDR_DRIVE_BUFFER_MAX_SIZE);
         }
      }
//msg(MSG_INFO,"Exec start %x %x\n", pru_get_pc(0), pru_get_pc(1));
      pru_exec_cmd(CMD_START, 0);
//msg(MSG_INFO,"Exec start done\n");
      seek_time = pru_read_word(MEM_PRU0_DATA,PRU0_SEEK_TIME) / 200e6 * 1e3;
      if (seek_time > max_seek_time) {
         max_seek_time = seek_time;
      }
      msg(MSG_INFO,"  Waiting, seek time %.1f ms max %.1f\n", seek_time, max_seek_time);
      // Wait for request from PRU
      prussdrv_pru_wait_event (PRU_EVTOUT_0);
      prussdrv_pru_clear_event (PRU_EVTOUT_0, PRU0_ARM_INTERRUPT);
      for (i = 0; i < drive_params->num_drives; i++) {
         dirty = pru_read_word(MEM_PRU1_DATA, PRU1_DRIVE0_TRK_DIRTY + i*PRU_WORD_SIZE_BYTES);
         new_cyl[i] = pru_read_word(MEM_PRU0_DATA,PRU0_DRIVE0_CUR_CYL +  i*PRU_WORD_SIZE_BYTES);
         if (new_cyl[i] == -1) {
            done = 1;
         }
         b = pru_read_word(MEM_PRU0_DATA,PRU0_CUR_SELECT_HEAD);
         if (cyl[i] != new_cyl[i]) {
            msg(MSG_INFO,"  Drive %d Cyl %d,%d select %d, head %d dirty %x\n", i, cyl[i], new_cyl[i],
               (((b >> 22) & 0x3) | ((b >> 24) & 0xc)) ^ 0xf, 
               ((b >> 2) & 0xff) ^ 0xf, dirty);
         }
         if (dirty) {
            pru_write_word(MEM_PRU1_DATA,PRU1_DRIVE0_TRK_DIRTY +
                  i*PRU_WORD_SIZE_BYTES, 0);
            pru_read_mem(MEM_DDR, data[i], cyl_size[i], i*DDR_DRIVE_BUFFER_MAX_SIZE);
            emu_file_write_cyl(drive_params->fd[i], cyl[i], data[i], cyl_size[i]);
         }
      }
   }
   // Tell PRU we saw the exiting flag
   pru_write_word(MEM_PRU0_DATA,PRU0_DRIVE0_CUR_CYL, 0);

   // Done with file
   emu_file_close(drive_params->fd[i], 0);

   return NULL;
}

// Main routine. 
int main(int argc, char *argv[])
{
   // Size of shared memory in bytes
   int ddr_mem_size;
   int i;
   DRIVE_PARAMS drive_params;
   uint32_t zeros[EMU_FILE_TRACK_DATA_BYTES/4];
   int cyl,head;
   int select_map[] = 
      {0, GPIO_SELECT1, GPIO_SELECT2, GPIO_SELECT3, GPIO_SELECT4};

   memset(zeros, 0, sizeof(zeros));
   // Find out what we should do
   parse_cmdline(argc, argv, &drive_params);

   if (drive_params.initialize) {
      if (drive_params.num_drives != 1) {
         msg(MSG_FATAL, "Only one filename may be specified when initializing\n");
         exit(1);
      }
      drive_params.fd[0] = emu_file_write_header(drive_params.filename[0],
        drive_params.num_cyl, drive_params.num_head);
      for (cyl = 0; cyl < drive_params.num_cyl; cyl++) {
         for (head = 0; head < drive_params.num_head; head++) {
            emu_file_write_track_bits(drive_params.fd[0], zeros,
                  EMU_FILE_TRACK_DATA_BYTES/4, cyl, head);
         }
      }
      emu_file_close(drive_params.fd[0], 1);
   }
   // Initialize PRU
   ddr_mem_size = pru_setup(2);
   if (ddr_mem_size == -1) {
      exit(1);
   }

   for (i = 0; i < MAX_DRIVES; i++) {
      // 0 for drive always selected, 1-4 for select 1-4
      if (i < drive_params.num_drives) {
         if (drive_params.drive[i] < ARRAYSIZE(select_map)) {
            pru_write_word(MEM_PRU0_DATA,PRU0_DRIVE0_SELECT +
                  i*PRU_WORD_SIZE_BYTES,
                  select_map[drive_params.drive[i]]);
         } else {
            msg(MSG_FATAL, "Illegal drive select %d\n", drive_params.drive);
            exit(1);
         }
      } else {
         pru_write_word(MEM_PRU0_DATA,PRU0_DRIVE0_SELECT + i*
               (PRU0_DRIVE1_SELECT - PRU0_DRIVE0_SELECT), 0);
      }
   }
   // And start our code
   if (prussdrv_exec_program(0, "prucode0.bin") != 0) {
      msg(MSG_FATAL, "Unable to execute prucode0.bin\n");
      exit(1);
   }

   if (prussdrv_exec_program(1, "prucode1.bin") != 0) {
      msg(MSG_FATAL, "Unable to execute prucode1.bin\n");
      exit(1);
   }

   // this messes up IEP running/accessing registers
   // Cleanup when we exit
   signal(SIGINT,(__sighandler_t) shutdown_signal);
   atexit(shutdown);

   if (pthread_create(&read_thread, NULL, &emu_proc, &drive_params)
         != 0) {
      msg(MSG_FATAL, "Unable to create read thread\n");
      exit(1);
   }

#if 1
   // TODO: Should this drop speed when idle?
   // This is needed to keep up with the data. Without it we will take
   // more than 2 revolutions per track due to the default governor not
   // increasing the CPU speed enough. We switch frequently between busy and
   // sleeping.
   if (set_restore_max_cpu_speed(0)) {
      msg(MSG_ERR, "Unable to set CPU to maximum speed\n");
   }
#endif
   sleep(1);
// This is debugging stuff to print state of data from PRU
struct {
   int loc;
   MEM_TYPE mem_type;
   char *txt;
   uint32_t last;
} locs[] = {
   {PRU1_BAD_PATTERN_COUNT, MEM_PRU1_DATA, "bad pattern count %d\n", 123},
   {PRU0_RQUEUE_UNDERRUN, MEM_PRU0_DATA, "Read queue underrun %d\n", 123},
   {PRU0_WQUEUE_OVERRUN, MEM_PRU0_DATA, "Write queue overrun %d\n", 123},
   {PRU0_ECAP_OVERRUN, MEM_PRU0_DATA, "Ecapture overrun %d\n", 123},
   {PRU0_HEAD_SELECT_GLITCH_COUNT, MEM_PRU0_DATA, "glitch count %d\n", 123},
   {PRU0_HEAD_SELECT_GLITCH_VALUE, MEM_PRU0_DATA, "glitch value %x\n", 123},
   {PRU_TEST0, MEM_PRU0_DATA, "0:test 0 %x\n", 123},
   {PRU_TEST1, MEM_PRU0_DATA, "0:test 1 %x\n", 123},
   {PRU_TEST2, MEM_PRU0_DATA, "0:test 2 %x\n", 123},
   {PRU_TEST3, MEM_PRU0_DATA, "0:test 3 %x\n", 123},
   {PRU_TEST4, MEM_PRU0_DATA, "0:test 4 %x\n", 123},
   {PRU_TEST0, MEM_PRU1_DATA, "1:test 0 %x\n", 123},
   {PRU_TEST1, MEM_PRU1_DATA, "1:test 1 %x\n", 123},
   {PRU_TEST2, MEM_PRU1_DATA, "1:test 2 %x\n", 123},
   {PRU_TEST3, MEM_PRU1_DATA, "1:test 3 %x\n", 123},
   {PRU_TEST4, MEM_PRU1_DATA, "1:test 4 %x\n", 123}
};
   while (1) {
      uint32_t a;
#if 1
      for (i = 0; i < ARRAYSIZE(locs); i++) {
         a = pru_read_word(locs[i].mem_type,locs[i].loc);
         if (a != locs[i].last) {
            printf(locs[i].txt,a);
            locs[i].last = a;
         }
      } 
#endif
#if 1
static uint32_t b, last_b = 9;
      b = pru_read_word(MEM_PRU0_DATA,PRU0_CUR_SELECT_HEAD);
      if (b != last_b) {
         printf("select %d head %d\n",
               (((b >> 22) & 0x3) | ((b >> 24) & 0xc)) ^ 0xf, 
               ((b >> 2) & 0xff) ^ 0xf);
         last_b = b;
      } 
#endif
      //printf("PC %x %x\n", pru_get_pc(0), pru_get_pc(1));
      usleep(1000000);
   }

   // Not reached, exit called to terminate
   return (0);
}

